@@ -1,6 +1,6 @@
 package BigIP::GTM::ParseConfig;
 
-# CURRENTLY UNDER DEVELOMENT BY WENWU YAN < careline@126.com >
+# CURRENTLY UNDER DEVELOMENT BY WENWU YAN
 #----------------------------------------------------------------------------
 # The contents of this file are subject to the iControl Public License
 # Version 4.5 (the "License"); you may not use this file except in
@@ -16,7 +16,7 @@ package BigIP::GTM::ParseConfig;
 # distributed by F5.
 #
 # The Initial Developer of the Original Code is F5 Networks,
-# Inc. Seattle, WA, USA. Portions created by F5 are Copyright (C) 1996-2019 F5 Networks,
+# Inc. Seattle, WA, USA. Portions created by F5 are Copyright (C) 1996-2020 F5 Networks,
 # Inc. All Rights Reserved.  iControl (TM) is a registered trademark of F5 Networks, Inc.
 #
 # Alternatively, the contents of this file may be used under the terms
@@ -30,7 +30,10 @@ package BigIP::GTM::ParseConfig;
 # version of this file under either the License or the GPL.
 #----------------------------------------------------------------------------
 
-our $VERSION = '0.8.2';
+#----------------------------------------------------------------------------
+# Initialize the module
+#----------------------------------------------------------------------------
+our $VERSION = '0.83';
 my $AUTOLOAD;
 
 use 5.012;
@@ -50,7 +53,9 @@ sub new {
     return $self;
 }
 
-# Return a list of objects
+#----------------------------------------------------------------------------
+# Return a list of objects for ltm
+#----------------------------------------------------------------------------
 sub regions     { return shift->_objectlist('gtm region'); }
 sub wideips     { return shift->_objectlist('gtm wideip'); }
 sub pools       { return shift->_objectlist('gtm pool'); }
@@ -65,7 +70,9 @@ sub interfaces  { return shift->_objectlist('net interface'); }
 sub mgmt_routes { return shift->_objectlist('sys management-route'); }
 sub users       { return shift->_objectlist('auth user'); }
 
-# Return an object hash
+#----------------------------------------------------------------------------
+# Return a list of objects for net
+#----------------------------------------------------------------------------
 sub region     { return shift->_object( 'gtm region',           shift ); }
 sub wideip     { return shift->_object( 'gtm wideip',           shift ); }
 sub pool       { return shift->_object( 'gtm pool',             shift ); }
@@ -84,8 +91,9 @@ sub ntp        { return shift->_object( 'sys',                  'ntp' ); }
 sub syslog     { return shift->_object( 'sys',                  'syslog' ); }
 sub user       { return shift->_object( 'auth user',            shift ); }
 
-#return a list of funcs
-
+#----------------------------------------------------------------------------
+# Return a list of objects for others
+#----------------------------------------------------------------------------
 sub funcs {
     my $self = shift;
 
@@ -116,7 +124,9 @@ sub members {
     }
 }
 
-#Return a list of VS
+#----------------------------------------------------------------------------
+# Return a list of wideips_all
+#----------------------------------------------------------------------------
 sub wideips_all {
     my $self = shift;
     $self->{'Parsed'} ||= $self->_parse();
@@ -125,28 +135,81 @@ sub wideips_all {
 
     # Loop for wideips()
     foreach ( $self->wideips() ) {
-        my $pools = $self->wideip($_)->{pools};
+        my $ret = $self->wideip($_);
+        my $pools = $ret->{pools};
 
         foreach my $pool ( keys %{$pools} ) {
-            my $members = $self->pool($pool)->{members}
-                if defined $self->pool($pool);
-
-            $self->wideip($_)->{"pools_members"} = $members;
+            my $detail = $self->pool($pool);
+            my $members = $detail->{members} if defined $detail->{members};
+            $ret->{"detail"}{$pool}{"pool_detail"} = $detail;
 
             foreach my $serverAndVs ( keys %{$members} ) {
                 my ( $server, $vs ) = split( /\:/, $serverAndVs );
-                my $details = $self->server($server);
+                my $detail         = $self->server($server);
+                my $virtual_server  = $detail->{"virtual-servers"}{$vs} if $detail;
+                $ret->{"detail"}{$pool}{"server_detail"} ||= [];
+                my $mem = $detail if $detail;
+                $mem->{"server"} = $server;
+                $mem->{"vs"} = $vs;
+                push @{$ret->{"detail"}{$pool}{"server_detail"}}, $mem;
 
-                $self->wideip($_)->{"servers_details"}{$serverAndVs} = $server
-                    if $details;
-                $self->wideip($_)->{"vs_details"}{$serverAndVs}
-                    = $details->{"virtual-servers"}{$vs}
-                    if $vs;
+                my $monitor = $virtual_server->{"monitor"};
+                if ( $monitor && $monitor =~ /http|tcp|udp|bigip|gateway_icmp/ ) {
+                    next;
+                }
+                elsif ( $monitor && $monitor =~ /\S+/ ) {
+                    my $mon_detail = $self->monitor($monitor);
+                    $ret->{"detail"}{$pool}{"monitor_detail"} = $mon_detail if $mon_detail; 
+                }                
             }
         }
     }
 
-    return $self->{Parsed}{"gtm wideip"};
+    return $self->{"Parsed"}{"gtm wideip"};
+}
+
+#----------------------------------------------------------------------------
+# Return a list of wideip_detail
+#----------------------------------------------------------------------------
+sub wideip_detail {
+    my $self   = shift;
+    my $wideip = shift;
+
+    $self->{'Parsed'} ||= $self->_parse();
+
+    return 0 unless $self->{'Parsed'}->{'gtm wideip'};
+
+    # Loop for wideips()
+    my $ret = $self->wideip($wideip);
+    my $pools = $ret->{pools};
+
+    foreach my $pool ( keys %{$pools} ) {
+        my $detail = $self->pool($pool);
+        my $members = $detail->{members} if defined $detail->{members};
+        $ret->{"detail"}{$pool}{"pool_detail"} = $detail;
+
+        foreach my $serverAndVs ( keys %{$members} ) {
+            my ( $server, $vs ) = split( /\:/, $serverAndVs );
+            my $detail         = $self->server($server);
+            my $virtual_server  = $detail->{"virtual-servers"}{$vs} if $detail;
+            $ret->{"detail"}{$pool}{"server_detail"} ||= [];
+            my $mem = $detail if $detail;
+            $mem->{"server"} = $server;
+            $mem->{"vs"} = $vs;
+            push @{$ret->{"detail"}{$pool}{"server_detail"}}, $mem;
+
+            my $monitor = $virtual_server->{"monitor"};
+            if ( $monitor && $monitor =~ /http|tcp|udp|bigip|gateway_icmp/ ) {
+                next;
+            }
+            elsif ( $monitor && $monitor =~ /\S+/ ) {
+                my $mon_detail = $self->monitor($monitor);
+                $ret->{"detail"}{$pool}{"monitor_detail"} = $mon_detail if $mon_detail; 
+            }                
+        }
+    }
+
+    return $self->{Parsed}{"gtm wideip"}{$wideip};
 }
 
 # Modify an object
@@ -306,26 +369,26 @@ sub _parse {
     my @file = <FILE>;
     close FILE;
 
-    my ( $parsed, $obj, $key, $attr, $attr1, $attr2 );
+    my ( $parsed, $obj, $key, $attr1, $attr2, $attr3 );
 
     until ( !$file[0] ) {
         my $ln = shift @file;
 
         #policy hit situation with gtm attribute
         if ( $ln =~ /^(auth user|patition|cli)\s(.*)\s\{(\s\})?$/ ) {
-            $obj  = $1;
-            $key  = $2;
-            $attr = undef if $3;
+            next if $3;
+            $obj = $1;
+            $key = $2;
         }
 
         #gtm attribute
         elsif ( $ln
-            =~ /^(gtm wideip|gtm pool|gtm server|gtm monitor|gtm region)\s(.*)\s\{(\s\})?$/
+            =~ /^(gtm wideip|gtm pool|gtm server|gtm region)\s(.*)\s\{(\s\})?$/
             )
         {
-            $obj  = $1;
-            $key  = $2;
-            $attr = undef if $3;
+            next if $3;
+            $obj = $1;
+            $key = $2;
         }
 
         #net attribute
@@ -333,25 +396,36 @@ sub _parse {
             =~ /^(net self|net route|net interface|net vlan|net trunk)\s(.*)\s\{(\s\})?$/
             )
         {
-            $obj  = $1;
-            $key  = $2;
-            $attr = undef if $3;
+            next if $3;
+            $obj = $1;
+            $key = $2;
         }
 
         #sys attribute
         elsif ( $ln =~ /^(sys management-route)\s(.*)\s\{(\s\})?$/ ) {
-            $obj  = $1;
-            $key  = $2;
-            $attr = undef if $3;
+            next if $3;
+            $obj = $1;
+            $key = $2;
+        }
+
+        #gtm monitor attribute
+        elsif ( $ln
+            =~ /^((gtm monitor)\s(http|tcp|udp|bigip|gateway_icmp))\s(\S+)\s\{(\s\})?$/
+            )
+        {
+            next if $5;
+            $obj = $2;
+            $key = $4;
+            $parsed->{$obj}{$key}{"monitor_method"} = $3 if ( $obj && $key );
         }
 
         #sys management attribute
         elsif ( $ln
             =~ /^(sys)\s(snmp|sshd|ntp|syslog|state-mirroring)\s\{(\s\})?$/ )
         {
-            $obj  = $1;
-            $key  = $2;
-            $attr = undef if $3;
+            next if $3;
+            $obj = $1;
+            $key = $2;
         }
         elsif ( $ln =~ /^\}$/ ) {
             $obj = undef;
@@ -363,17 +437,23 @@ sub _parse {
             $self->{'Raw'}->{$obj}{$key} .= $ln;
 
             #Indent=4 { not empty }
-            if ( $ln =~ /^\s{4}(\S+)\s\{$/ ) {
+            if ( $ln =~ /^\s{4}(\S+|\".*\")\s\{$/ ) {
                 $attr1 = $1;
                 next;
             }
 
             #$intdent=8 { not empty }
-            if ( $ln =~ /^\s{8}(\S+)\s\{$/ ) {
+            if ( $ln =~ /^\s{8}(\S+|\".*\")\s\{$/ ) {
                 $attr2 = $1;
                 next;
             }
 
+            #$intdent=12 { not empty }
+            if ( $ln =~ /^\s{12}(\S+|\".*\")\s\{$/ ) {
+                $attr3 = $1;
+                next;
+            }
+
             #Indent=4 with }$
             if ( $ln =~ /^\s{4}\}$/ ) {
                 $attr1 = undef;
@@ -386,6 +466,12 @@ sub _parse {
                 next;
             }
 
+            #Indent=12 with }$
+            if ( $ln =~ /^\s{12}\}$/ ) {
+                $attr3 = undef;
+                next;
+            }
+
             #Indent=4 {}
             if ( $ln =~ /^\s{4}(\S+)\s\{\s\}$/ ) {
                 $parsed->{$obj}{$key}{$1} = undef;
@@ -400,7 +486,7 @@ sub _parse {
                 next;
             }
 
-            if ($attr1) {
+            if ( defined $attr1 && $attr1 ) {
 
                 #Indent=8 {}
                 if ( $ln =~ /^\s{8}(\S+)\s\{\s\}$/ ) {
@@ -410,8 +496,8 @@ sub _parse {
 
                 #Indent=8 { scalar }
                 if ( $ln =~ /^\s{8}(\S+)\s\{(.*)\}$/ ) {
-                    $parsed->{$obj}{$key}{$1} ||= [];
-                    push @{ $parsed->{$obj}{$key}{$1} },
+                    $parsed->{$obj}{$key}{$attr1}{$1} ||= [];
+                    push @{ $parsed->{$obj}{$key}{$attr1}{$1} },
                         grep { not /^\s*$/ } split( /\s+/, $2 );
                     next;
                 }
@@ -423,7 +509,7 @@ sub _parse {
                 }
 
                 #Indent=8 match scalar
-                if ( $ln =~ /^\s{8}(\S.+)$/ ) {
+                if ( $ln =~ /^\s{8}(\S+)$/ ) {
                     if ( ref( $parsed->{$obj}{$key}{$attr1} ) eq 'HASH' ) {
                         $parsed->{$obj}{$key}{$attr1}{$1} = undef;
                     }
@@ -435,7 +521,7 @@ sub _parse {
                 }
             }
 
-            if ($attr2) {
+            if ( defined $attr2 && ( $attr1 && $attr2 ) ) {
 
                 #Indent=12 match { not empty }
                 if ( $ln =~ /^\s{12}(\S+)\s\{\s\}$/ ) {
@@ -453,19 +539,59 @@ sub _parse {
 
                 #Indent=12 match { key => value }
                 if ( $ln =~ /^\s{12}(\S+)\s(.*)\s?$/ ) {
-                    say( $ln, $attr1, $attr2 );
                     $parsed->{$obj}->{$key}{$attr1}{$attr2}{$1} = $2;
                     next;
                 }
 
                 #Indent=12 match scalar
                 if ( $ln =~ /^\s{12}(.*)\s?$/ ) {
-                    if ( ref( $parsed->{$obj}{$key}{$attr1} ) eq 'HASH' ) {
-                        $parsed->{$obj}{$key}{$attr1}{$1} = undef;
+                    if (ref( $parsed->{$obj}{$key}{$attr1}{$attr2} ) eq
+                        'HASH' )
+                    {
+                        $parsed->{$obj}{$key}{$attr1}{$attr2}{$1} = undef;
                     }
                     else {
-                        $parsed->{$obj}{$key}{$attr1} ||= [];
-                        push @{ $parsed->{$obj}{$key}{$attr1} }, $1;
+                        $parsed->{$obj}{$key}{$attr1}{$attr2} ||= [];
+                        push @{ $parsed->{$obj}{$key}{$attr1}{$attr2} }, $1;
+                    }
+                    next;
+                }
+            }
+
+            if ( defined $attr3 && ( $attr1 && $attr2 && $attr3 ) ) {
+
+                #Indent=16 match { not empty }
+                if ( $ln =~ /^\s{16}(\S+)\s\{\s\}$/ ) {
+                    $parsed->{$obj}{$key}{$attr1}{$attr2}{$attr3}{$1} = undef;
+                    next;
+                }
+
+                #Indent=16 { scalar }
+                if ( $ln =~ /^\s{16}(\S+)\s\{(.*)\}$/ ) {
+                    $parsed->{$obj}{$key}{$attr1} ||= [];
+                    push @{ $parsed->{$obj}{$key}{$attr1}{$attr2}{$attr3}{$1}
+                    }, grep { not /^\s*$/ } split( /\s+/, $2 );
+                    next;
+                }
+
+                #Indent=16 match { key => value }
+                if ( $ln =~ /^\s{16}(\S+)\s(.*)\s?$/ ) {
+                    $parsed->{$obj}->{$key}{$attr1}{$attr2}{$attr3}{$1} = $2;
+                    next;
+                }
+
+                #Indent=12 match scalar
+                if ( $ln =~ /^\s{16}(.*)\s?$/ ) {
+                    if (ref( $parsed->{$obj}{$key}{$attr1}{$attr2}{$attr3} )
+                        eq 'HASH' )
+                    {
+                        $parsed->{$obj}{$key}{$attr1}{$attr2}{$attr3}{$1}
+                            = undef;
+                    }
+                    else {
+                        $parsed->{$obj}{$key}{$attr1}{$attr2}{$attr3} ||= [];
+                        push @{ $parsed->{$obj}{$key}{$attr1}{$attr2}{$attr3}
+                        }, $1;
                     }
                     next;
                 }
@@ -476,9 +602,6 @@ sub _parse {
                 $parsed->{$obj}{$key}{$1} = $2;
                 next;
             }
-            else {
-                say $ln;
-            }
         }
     }
 
@@ -495,18 +618,25 @@ sub _parse {
 1;
 
 __END__
-
+ 
+Hide 9 lines of Pod
 =head1 NAME
  
-BigIP::GTM::ParseConfig - F5/BigIP GTM configuration parser
+BigIP::LTM::ParseConfig - The great new BigIP::LTM::ParseConfig!
  
+=head1 VERSION
+ 
+Version 0.83
+ 
+=cut
+ 
+Hide 218 lines of Pod
 =head1 SYNOPSIS
- 
-  use BigIP::GTM::ParseConfig;
-   
+  use BigIP::LTM::ParseConfig;
+     
   # Module initialization
-  my $bip = new BigIP::GTM::ParseConfig( '/config/bigip.conf' );
-   
+  my $bip = new BigIP::LTM::ParseConfig( '/config/bigip.conf' );
+     
   # Iterate over pools
   foreach my $pool ( $bip->pools() ) {
       # Iterate over pool members
@@ -526,150 +656,196 @@ BigIP::GTM::ParseConfig - F5/BigIP GTM configuration parser
           );
       }
   }
-   
+     
   # Write out a new config file
   $bip->write( '/config/bigip.conf.new' );
- 
+   
 =head1 DESCRIPTION
- 
-BigIP::GTM::ParseConfig provides a Perl interface to reading, writing, and
+   
+BigIP::LTM::ParseConfig provides a Perl interface to reading, writing, and
 manipulating configuration files used on F5 (BigIP) LTM network devices.
- 
+   
 This module is currently a work-in-progress.  Please e-mail with problems,
 bug fixes, comments and complaints.
- 
+   
 =head1 CONSTRUCTOR
- 
+   
 =over 4
- 
+   
 =item new ( FILE )
- 
-Create a new B<BigIP::GTM::ParseConfig> object.
- 
+   
+Create a new B<BigIP::LTM::ParseConfig> object.
+   
 C<FILE> refers to the bigip.conf configuration file, usually found at
 /config/bigip.conf.
- 
+   
 B<Example>
- 
-  $bip = BigIP::GTM::ParseConfig->new( '/config/bigip.conf' );
- 
+   
+  $bip = BigIP::LTM::ParseConfig->new( '/config/bigip.conf' );
+   
 =back
- 
+   
 =head1 METHODS
- 
+   
 =over 4
- 
+   
 =item monitors
- 
+ @monitos = $bip->monitors;
 =item nodes
- 
+ @nodes = $bip->nodes;
 =item partitions
- 
+ @partions = $bip->partitions;
 =item pools
- 
+ @pools = $bip->pools;
 =item profiles
- 
+ @profiles = $bip->profiles;
 =item routes
- 
+ @routes = $bip->routes;
 =item rules
- 
+ @rules = $bip->rules;
 =item users
- 
+ @users = $bip->uers;
 =item virtuals
+ @virtuals = $bip->virtuals;
  
 List the names of all found objects of the referring method.
- 
+   
 B<Examples>
- 
+   
   @pools = $bip->pools();
- 
+   
   @virtuals = $bip->virtuals();
- 
+   
 =item monitor ( MONITOR )
- 
+   
 =item node ( NODE )
- 
+   
 =item partition ( PARTITION )
- 
+   
 =item pool ( POOL )
- 
+   
 =item profile ( PROFILE )
- 
+   
 =item route ( ROUTE )
- 
+   
 =item rule ( RULE )
- 
+   
 =item user ( USER )
- 
+   
 =item virtual ( VIRTUAL )
- 
+   
 Return a hash of the object specified.
- 
+   
 B<Examples>
- 
+   
   %sschneid = $bip->user( 'sschneid' );
- 
+   
   $monitor = $bip->pool( 'Production_LDAP_pool')->{'monitor'};
- 
+   
 =item members ( POOL )
- 
+   
 List the members of a specified pool.
- 
+   
 B<Example>
- 
+   
   @members = $bip->members( 'Production_LDAP_pool' );
- 
+   
 Note that this is identical to using the I<pool> method:
- 
+   
   @members = @{$bip->pool( 'Production_LDAP_pool' )->{'members'}};
- 
+   
 =item modify ( OPTIONS )
- 
+   
 Modify the attributes of a specified object.  The following options are B<required>:
- 
+   
 =over 4
- 
+   
 =item type
- 
+   
 The type of object being modified.  Allowed types are: monitor, node, partition,
 pool, profile, route, user, virtual.
- 
+   
 =item key
- 
+   
 The key (name) of the object being modified.
- 
+   
 =back
- 
+   
 Following B<type> and B<key> should be a string or a reference to an array of
 strings.  See the example below for more details.
- 
+   
 B<Examples>
- 
+   
   $bip->modify(
       type => 'virtual',
       key  => 'Production_LDAP_vip',
       persist => 'cookie'
   );
-   
+     
   $bip->modify(
       type => 'pool',
       key  => 'Production_LDAP_pool',
       members => [ '192.168.0.1:636', '192.168.0.2:636' ]
   );
- 
+   
 =item write ( FILE )
- 
+   
 Write out a new configuration file.  C<FILE> refers to the bigip.conf configuration
 file, usually found at /config/bigip.conf.
- 
+   
 B<Example>
- 
+   
   $bip->write( '/config/bigip.conf.new' );
+   
+=head1 AUTHOR
+ 
+WENWU YAN, C<< <careline at 126.com> >>
+ 
+=head1 BUGS
+ 
+Please report any bugs or feature requests to C<bug-bigip-ltm-parseconfig at rt.cpan.org>, or through
+the web interface at L<https://rt.cpan.org/NoAuth/ReportBug.html?Queue=BigIP-LTM-ParseConfig>.  I will be notified, and then you'll
+automatically be notified of progress on your bug as I make changes.
+ 
+=head1 SUPPORT
+ 
+You can find documentation for this module with the perldoc command.
+ 
+    perldoc BigIP::LTM::ParseConfig
+ 
+You can also look for information at:
+ 
+=over 4
+ 
+=item * RT: CPAN's request tracker (report bugs here)
+ 
+L<https://rt.cpan.org/NoAuth/Bugs.html?Dist=BigIP-LTM-ParseConfig>
+ 
+=item * AnnoCPAN: Annotated CPAN documentation
+ 
+L<http://annocpan.org/dist/BigIP-LTM-ParseConfig>
+ 
+=item * CPAN Ratings
+ 
+L<https://cpanratings.perl.org/d/BigIP-LTM-ParseConfig>
+ 
+=item * Search CPAN
+ 
+L<https://metacpan.org/release/BigIP-LTM-ParseConfig>
  
 =back
  
-=head1 AUTHOR
  
-WENWU YAN <careline@126.com>
+=head1 ACKNOWLEDGEMENTS
  
+ 
+=head1 LICENSE AND COPYRIGHT
+ 
+This software is Copyright (c) 2019 by WENWU YAN.
+ 
+This is free software, licensed under:
+ 
+  The Artistic License 2.0 (GPL Compatible)
+   
 =cut
+
